//------------------------------------------------------------------------------
// Copyright (C) 2010, Raditex AB
// All rights reserved.
//
// FreeSCADA 
// http://www.FreeSCADA.com
// freescada@freescada.com
//
//------------------------------------------------------------------------------

#include "net.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/time.h>
#include <time.h>

#include <glib.h>

//------------------------------------------------------------------------------
// Close down a socket.
//------------------------------------------------------------------------------
int
fs_net_socket_close(int sock)
{
	static const char func_name[] = "fs_net_socket_close";
	g_message("%s: close socket: %d", func_name, sock);
	close(sock);
	return -1;
}

//------------------------------------------------------------------------------
// Establish a TCP/IP connection to a server with hostname and port.
// Set up callback functions (given by io_cb) that is supposed to be
// called when data is available from the server, or a socket event
// such as disconnect etc. happens.
//------------------------------------------------------------------------------
GIOChannel *
fs_net_connection_setup(char *hostname, int port, void *io_cb, gpointer cb_data)
{
   static const char func_name[] = "fs_net_connection_setup";
   struct hostent *host_addr;
   struct sockaddr_in s_addr;
   int sock;
   struct timeval time_out;

   GIOChannel *gio_sock;

   if ((sock = socket(AF_INET,SOCK_STREAM, 0)) < 0)
   {
      g_warning("%s: failed to setup a socket.", func_name);
      return NULL;
   }

   s_addr.sin_family = AF_INET;
   s_addr.sin_port = htons(port);

   /* resolve hostname */
   if ((host_addr = gethostbyname(hostname)) == NULL)
   {
      g_warning("%s: unknown host: %s", func_name, hostname);
      return NULL;
   }
   bcopy((char *) (host_addr->h_addr), (char *) (&s_addr.sin_addr),
         host_addr->h_length);

   if (connect(sock, (struct sockaddr *) &s_addr, sizeof(s_addr)) < 0)
   {
      g_warning("%s: Failed to establish connection to %s:%d", func_name,
            hostname, port);
      return NULL;
   }

   // Try to set a timeout...
   time_out.tv_sec = 5;
   time_out.tv_usec = 0;
   setsockopt(sock, SOL_SOCKET, SO_SNDTIMEO, &time_out, sizeof(time_out));
   setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, &time_out, sizeof(time_out));

   /* set up a GLIB IO Channel for the socket */
   if ((gio_sock = g_io_channel_unix_new(sock)) == NULL)
   {
      g_warning("%s: Couldn't create a new GIOChannel.", func_name);
      return NULL;
   }

   /* set up I/O event handler */
   if (!g_io_add_watch(gio_sock, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
         io_cb, cb_data))
   {
      g_warning("%s: Couldn't add watch on GIOChannel.", func_name);
      return NULL;
   }

   g_io_channel_set_encoding(gio_sock, NULL, NULL);
   g_io_channel_set_buffered(gio_sock, FALSE);

   return gio_sock;
}

//------------------------------------------------------------------------------
// Connect an already initialized socket to an GIO channel.
//------------------------------------------------------------------------------
GIOChannel *
fs_net_create_gio(int sock, void *io_cb, gpointer cb_data)
{
    GIOChannel *gio_sock;
   
    /* set up a GLIB IO Channel for the socket */
    if ((gio_sock = g_io_channel_unix_new(sock)) == NULL)
    {
        g_warning("%s: Couldn't create a new GIOChannel.", __PRETTY_FUNCTION__);
        return NULL;
    }

    /* set up I/O event handler */
    if (!g_io_add_watch(gio_sock, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL, io_cb, cb_data))
    {
        g_warning("%s: Couldn't add watch on GIOChannel.", __PRETTY_FUNCTION__);
        return NULL;
    }

    g_io_channel_set_encoding(gio_sock, NULL, NULL);
    g_io_channel_set_buffered(gio_sock, FALSE);
    
    return gio_sock;
}

//------------------------------------------------------------------------------
// Close down a socket and GIO Channel.
//------------------------------------------------------------------------------
gint
fs_net_disconnect(GIOChannel *gio)
{
	static const char func_name[] = "fs_net_disconnect";
	GError *err = NULL;
	int unix_fd;

	if (gio == NULL)
	{
		g_message("%s: gio = NULL no g_io_channel to shut down.", func_name);
		return -1;
	}

	unix_fd = g_io_channel_unix_get_fd(gio);

	g_io_channel_shutdown(gio, TRUE, &err);

	g_message("%s: shutdown g_io_channel.", func_name);
	close(unix_fd);

	return 0;
}

//------------------------------------------------------------------------------
// fs_net_send:
//
// This functions sends a packet on a given GIO Channel:
//------------------------------------------------------------------------------
gint
fs_net_send(GIOChannel *gio, gchar *buff, gint buff_len)
{
   static const char func_name[] = "fs_net_send";
   gint ret;
   gsize len;

   if ((ret = g_io_channel_write(gio, buff, buff_len, &len)) != G_IO_ERROR_NONE)
   {
      g_message("%s: Failed to write data to [%d].", func_name, (int) gio);
      return -1;
   }
   else
   {
      if (len != buff_len)
      {
         g_message("%s: Wrote %d bytes (out of %d requested) to [%d].",
               func_name, len, buff_len, (int) gio);
      }
   }

   return (gint) len;
}

//------------------------------------------------------------------------------
// fs_net_read_data_nonblocking
//
//------------------------------------------------------------------------------
gint
fs_net_read_data_nonblocking(GIOChannel *gio, FsNetBuffer *data, gint data_length)
{
    static const char func_name[] = "fs_net_read_data_nonblocking";
    GIOStatus ret;
    GError *gerror = NULL;
    gsize bytes_read;

    if (data->offset == 0)
    {
        memset(data->buff, 0, FS_NET_BUFFER_SIZE);
    }

    while (data->offset < FS_NET_BUFFER_SIZE && data->offset < data_length)
    {
        ret = g_io_channel_read_chars(gio, &(data->buff[data->offset]), 1, &bytes_read, &gerror);

        if (ret == G_IO_STATUS_AGAIN)
        {
            //
            // We use a special return value (-2) to mark that we ran out of 
            // non-blocking data. We return -2 to indicate to the caller that
            // the read was incomplete and should be restarted on the next
            // callback.
            return -2;
            //break;
        }

        if (ret != G_IO_STATUS_NORMAL)
        {
            // This happens when the socket is disconnected. return -1
            // to indicate to the caller that the socket is shut down.

            data->offset = 0;
            return -1;
        }

        if (bytes_read == 0)
        {
            g_message("%s: ran out of data, premature return [%d].", func_name, data->offset);
            break;
        }

        data->offset++;
 
        if (data->offset == data_length)
        {
            break;
        }
   }

    ret = data->offset;
    data->offset  =  0;
    return ret;
}


//------------------------------------------------------------------------------
// Data buffer management:
//
//
static GList *read_buffer_list = NULL;
static gint list_len = 0;
gint
fs_net_buffer_new(GIOChannel *gio)
{
    static gchar *func_name = "fs_net_buffer_new";
    FsNetBuffer *nbdata;

    if ((nbdata = fs_net_buffer_get(gio)) == NULL)
    {
        if ((nbdata = g_malloc(sizeof(FsNetBuffer))) == NULL)
        {
            g_warning("%s: Failed to allocate memory for data buffer.", func_name);
            return -1;
        }

        memset((void *)nbdata, 0, sizeof(FsNetBuffer));
        nbdata->gio     =  gio;
        nbdata->offset  =  0;
        nbdata->tag_idx = -1;

        read_buffer_list = g_list_append(read_buffer_list, (gpointer)nbdata);
        list_len++;

    }
    else
    {
        g_warning("%s: A data buffer is already associated with GIO channel %d [%d].",
                  func_name, (gint)gio, list_len);
        return -1;
    }

    return 0;
}

void
fs_net_buffer_free(GIOChannel *gio)
{
    static gchar *func_name = "fs_net_buffer_free";
    FsNetBuffer *nbdata;

    if ((nbdata = fs_net_buffer_get(gio)) == NULL)
    {
        g_warning("%s: No data buffer is associated with the GIO channel %d.",
                  func_name, (gint)gio);
        return;        
    }

    read_buffer_list = g_list_remove(read_buffer_list, nbdata);
    g_free(nbdata);
    list_len--;

    return;
}


FsNetBuffer *
fs_net_buffer_get(GIOChannel *gio)
{
    FsNetBuffer *nbdata = NULL;
    GList *item;

    for (item = read_buffer_list; item; item = g_list_next(item))
    {
        nbdata = (FsNetBuffer *)item->data;

        if (nbdata && nbdata->gio == gio)
            return nbdata;
    }

    return NULL;
}


