//------------------------------------------------------------------------------
// Copyright (C) 2010, Raditex AB
// All rights reserved.
//
// FreeSCADA 
// http://www.FreeSCADA.com
// freescada@freescada.com
//
//------------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>

#include "mbus-protocol.h"

#define NITEMS(x) (sizeof(x)/sizeof(x[0]))

//------------------------------------------------------------------------------
// internal data
//------------------------------------------------------------------------------
static mbus_slave_data slave_data[MBUS_MAX_PRIMARY_SLAVES];

//------------------------------------------------------------------------------
// Return a pointer to the slave_data register. This register can be used for 
// storing current slave status.
//------------------------------------------------------------------------------
mbus_slave_data *
mbus_slave_data_get(int i)
{
    if (i < MBUS_MAX_PRIMARY_SLAVES)
    {
        return &slave_data[i];
    }
    
    return NULL;
}

//------------------------------------------------------------------------------
//
// M-Bus FRAME RELATED FUNCTIONS
//
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Allocate an M-bus frame data structure and initialize it according to which
// frame type is requested.
//------------------------------------------------------------------------------
mbus_frame *
mbus_frame_new(int frame_type)
{
    mbus_frame *frame = NULL;
    
    if ((frame = malloc(sizeof(mbus_frame))) != NULL)
    {
        memset((void *)frame, 0, sizeof(mbus_frame));
        
        frame->type = frame_type;
        switch (frame->type)
        {
            case MBUS_FRAME_TYPE_ACK:
            
                frame->start1 = MBUS_FRAME_ACK_START;
            
                break;

            case MBUS_FRAME_TYPE_SHORT:
            
                frame->start1 = MBUS_FRAME_SHORT_START;
                frame->stop   = MBUS_FRAME_STOP;  

                break;

            case MBUS_FRAME_TYPE_CONTROL:

                frame->start1 = MBUS_FRAME_CONTROL_START;
                frame->start2 = MBUS_FRAME_CONTROL_START;
                frame->length1 = 3;
                frame->length2 = 3;
                frame->stop   = MBUS_FRAME_STOP;  
                  
                break;

            case MBUS_FRAME_TYPE_LONG:

                frame->start1 = MBUS_FRAME_LONG_START;
                frame->start2 = MBUS_FRAME_LONG_START;
                frame->stop   = MBUS_FRAME_STOP;              

                break;
        }            
    }

    return frame;
}

//------------------------------------------------------------------------------
// Free the memory resources allocated for the M-Bus frame data structure.
//------------------------------------------------------------------------------
int
mbus_frame_free(mbus_frame *frame)
{
    if (frame)
    {
        free(frame);
        return 0;
    }
    return -1;
}

//------------------------------------------------------------------------------
// Caclulated the checksum of the M-Bus frame. The checksum algorithm is the
// arithmetic sum of the frame content, without using carry. Which content
// that is included in the checksum calculation depends on the frame type.
//------------------------------------------------------------------------------
int
mbus_frame_calc_checksum(mbus_frame *frame)
{
    size_t i;
    
    if (frame)
    {
        frame->checksum  = 0;
        
        switch (frame->type)
        {
            case MBUS_FRAME_TYPE_ACK:
            
                break;

            case MBUS_FRAME_TYPE_SHORT:
            
                frame->checksum  = 0;
                frame->checksum += frame->control;
                frame->checksum += frame->address;
            
                break;

            case MBUS_FRAME_TYPE_CONTROL:
            
                frame->checksum += frame->control;
                frame->checksum += frame->address;
                frame->checksum += frame->control_information;
        
                break;

            case MBUS_FRAME_TYPE_LONG:
            
                frame->checksum += frame->control;
                frame->checksum += frame->address;
                frame->checksum += frame->control_information;
        
                for (i = 0; i < frame->data_size; i++)
                {
                    frame->checksum += frame->data[i];
                }
            
                break;
        
            default:
                return -1;
        }        
    }
    
    return 0;
}

//------------------------------------------------------------------------------
// Calculate the values of the lengths fields in the M-Bus frame.
//------------------------------------------------------------------------------
int
mbus_frame_calc_length(mbus_frame *frame)
{
    if (frame)
    {
        frame->length1 = 0;
        frame->length2 = 0;
        
        switch (frame->type)
        {
            case MBUS_FRAME_TYPE_CONTROL:

                frame->length1 = 3;
                frame->length2 = 3;

                break;

            case MBUS_FRAME_TYPE_LONG:
           
                frame->length1 = frame->data_size + 3;
                frame->length2 = frame->data_size + 3;
            
                break;
        }        
    }
    
    return 0;
}

//------------------------------------------------------------------------------
// Return the M-Bus frame type
//------------------------------------------------------------------------------
int
mbus_frame_type(mbus_frame *frame)
{
    if (frame)
    {
        return frame->type;
    }
    return -1;
}

//------------------------------------------------------------------------------
// Verify that parsed frame is a valid M-bus frame. 
//
// Possible checks:
//
// 1) Start/stop bytes
// 2) length field and actual data size
// 3) checksum
//
//------------------------------------------------------------------------------
int
mbus_frame_verify(mbus_frame *frame)
{
    if (frame)
    {
        // for now let everything pass...
        return 0;
    }
    
    return -1;
}


//------------------------------------------------------------------------------
//
// DATA ENCODING, DECODING, AND CONVERSION FUNCTIONS
//
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Encode BCD data
//------------------------------------------------------------------------------
int
mbus_data_bcd_encode(u_char *bcd_data, size_t bcd_data_size, int value)
{ 
    int val = 0, v0, v1, v2, x1, x2;
    size_t i;

    v2 = value;

    if (bcd_data)
    {
        for (i = 0; i < bcd_data_size; i++)
        {
            v0 = v2;
            v1 = (int)(v0 / 10.0);
            v2 = (int)(v1 / 10.0);

            x1 = v0 - v1 * 10;
            x2 = v1 - v2 * 10;

            bcd_data[bcd_data_size-1-i] = (x2 << 4) | x1;
        }

        return 0;    
    }
    
    return -1;
}

//------------------------------------------------------------------------------
// Decode BCD data
//------------------------------------------------------------------------------
int
mbus_data_bcd_decode(u_char *bcd_data, size_t bcd_data_size)
{       
    int val = 0;
    size_t i;
    
    if (bcd_data)
    {
        //for (i = bcd_data_size-1; i >= 0; i--)
        for (i = 0; i < bcd_data_size; i++)
        {
            val = (val * 10) + ((bcd_data[i]>>4) & 0xF);
            val = (val * 10) + ( bcd_data[i]     & 0xF);
        }

        return val;    
    }
    
    return -1;
}

//------------------------------------------------------------------------------
// Decode INTEGER data
//------------------------------------------------------------------------------
int
mbus_data_int_decode(u_char *int_data, size_t int_data_size)
{
    int val = 0, i;
    
    if (int_data)
    {
        for (i = int_data_size-1; i >= 0; i--)
        {
            val = (val << 8) + (int_data[i]&0xFF);
        }

        return val;    
    }

    return -1;
}

//------------------------------------------------------------------------------
// Encode INTEGER data (into 'int_data_size' bytes)
//------------------------------------------------------------------------------
int
mbus_data_int_encode(u_char *int_data, size_t int_data_size, int value)
{
    int val = 0, i;
    
    if (int_data)
    {
        for (i = 0; i < int_data_size; i++)
        {
            int_data[i] = (value>>(i*8)) & 0xFF;
        }

        return 0;
    }

    return -1;
}

//------------------------------------------------------------------------------
// Generate manufacturer code from 2-byte encoded data
//------------------------------------------------------------------------------
int
mbus_data_manufacturer_encode(u_char *m_data, u_char *m_code)
{
    int m_val;

    if (m_data == NULL || m_code == NULL)
        return -1;

    m_val = ((((int)m_code[0] - 64) & 0x001F) << 10) + 
            ((((int)m_code[1] - 64) & 0x001F) << 5) +
            ((((int)m_code[2] - 64) & 0x001F));

    mbus_data_int_encode(m_data, 2, m_val);
    
    return 0;
}

//------------------------------------------------------------------------------
// Generate manufacturer code from 2-byte encoded data
//------------------------------------------------------------------------------
const char *
mbus_decode_manufacturer(u_char byte1, u_char byte2)
{
    static char m_str[4];
    
    int m_id;
    
    m_str[0] = byte1;
    m_str[1] = byte2;
    
    m_id = mbus_data_int_decode(m_str, 2);
        
    m_str[0] = (char)(((m_id>>10) & 0x001F) + 64);
    m_str[1] = (char)(((m_id>>5)  & 0x001F) + 64);
    m_str[2] = (char)(((m_id)     & 0x001F) + 64);
    m_str[3] = 0;

    return m_str;
}

//------------------------------------------------------------------------------
//
// FIXED-LENGTH DATA RECORD FUNCTIONS
//
//------------------------------------------------------------------------------


//
//   Value         Field Medium/Unit              Medium
// hexadecimal Bit 16  Bit 15    Bit 8  Bit 7
//     0        0       0         0     0         Other
//     1        0       0         0     1         Oil
//     2        0       0         1     0         Electricity
//     3        0       0         1     1         Gas
//     4        0       1         0     0         Heat
//     5        0       1         0     1         Steam
//     6        0       1         1     0         Hot Water
//     7        0       1         1     1         Water
//     8        1       0         0     0         H.C.A.
//     9        1       0         0     1         Reserved
//     A        1       0         1     0         Gas Mode 2
//     B        1       0         1     1         Heat Mode 2
//     C        1       1         0     0         Hot Water Mode 2
//     D        1       1         0     1         Water Mode 2
//     E        1       1         1     0         H.C.A. Mode 2
//     F        1       1         1     1         Reserved
//
const char *
mbus_data_fixed_medium(mbus_data_fixed *data)
{
    static char buff[256];

    if (data)
    {
        switch ( (data->cnt1_type&0xC0)>>4 | (data->cnt2_type&0xC0)>>6 )
        {
            case 0x00:    
                snprintf(buff, sizeof(buff), "Other");
                break;    
            case 0x01:    
                snprintf(buff, sizeof(buff), "Oil");
                break;    
            case 0x02:    
                snprintf(buff, sizeof(buff), "Electricity");
                break;    
            case 0x03:    
                snprintf(buff, sizeof(buff), "Gas");
                break;    
            case 0x04:    
                snprintf(buff, sizeof(buff), "Heat");
                break;    
            case 0x05:    
                snprintf(buff, sizeof(buff), "Steam");
                break;    
            case 0x06:    
                snprintf(buff, sizeof(buff), "Hot Water");
                break;    
            case 0x07:    
                snprintf(buff, sizeof(buff), "Water");
                break;    
            case 0x08:    
                snprintf(buff, sizeof(buff), "H.C.A.");
                break;    
            case 0x09:    
                snprintf(buff, sizeof(buff), "Reserved");
                break;    
            case 0x0A:    
                snprintf(buff, sizeof(buff), "Gas Mode 2");
                break;
            case 0x0B:
                snprintf(buff, sizeof(buff), "Heat Mode 2");
                break;
            case 0x0C:
                snprintf(buff, sizeof(buff), "Hot Water Mode 2");
                break;
            case 0x0D:
                snprintf(buff, sizeof(buff), "Water Mode 2");
                break;
            case 0x0E:
                snprintf(buff, sizeof(buff), "H.C.A. Mode 2");
                break;    
            case 0x0F:
                snprintf(buff, sizeof(buff), "Reserved");
                break;    
            default:
                snprintf(buff, sizeof(buff), "unknown");
                break;                
        }
        
        return buff;
    }

    return NULL;
}


//------------------------------------------------------------------------------
//                        Hex code                            Hex code
//Unit                    share     Unit                      share
//              MSB..LSB                            MSB..LSB
//                       Byte 7/8                            Byte 7/8
// h,m,s         000000     00        MJ/h           100000     20
// D,M,Y         000001     01        MJ/h * 10      100001     21
//     Wh        000010     02        MJ/h * 100     100010     22
//     Wh * 10   000011     03        GJ/h           100011     23
//     Wh * 100  000100     04        GJ/h * 10      100100     24
//   kWh         000101     05        GJ/h * 100     100101     25
//  kWh   * 10   000110     06           ml          100110     26
//   kWh * 100   000111     07           ml * 10     100111     27
//   MWh         001000     08           ml * 100    101000     28
//   MWh * 10    001001     09            l          101001     29
//   MWh * 100   001010     0A            l * 10     101010     2A
//     kJ        001011     0B            l * 100    101011     2B
//     kJ * 10   001100     0C           m3          101100     2C
//     kJ * 100  001101     0D        m3 * 10        101101     2D
//     MJ        001110     0E        m3 * 100       101110     2E
//     MJ * 10   001111     0F        ml/h           101111     2F
//     MJ * 100  010000     10        ml/h * 10      110000     30
//     GJ        010001     11        ml/h * 100     110001     31
//     GJ * 10   010010     12         l/h           110010     32
//     GJ * 100  010011     13         l/h * 10      110011     33
//      W        010100     14         l/h * 100     110100     34                              
//      W * 10   010101     15       m3/h           110101     35                         
//      W * 100  010110     16     m3/h * 10       110110     36
//     kW        010111     17      m3/h * 100       110111     37
//     kW * 10   011000     18        °C* 10-3       111000     38
//     kW * 100  011001     19      units   for HCA  111001     39
//     MW        011010     1A    reserved           111010     3A
//     MW * 10   011011     1B    reserved           111011     3B
//     MW * 100  011100     1C    reserved           111100     3C
//  kJ/h         011101     1D    reserved           111101     3D
//  kJ/h * 10    011110     1E    same but historic  111110     3E
//  kJ/h * 100   011111     1F    without   units    111111     3F
//
//------------------------------------------------------------------------------
const char *
mbus_data_fixed_unit(int medium_unit_byte)
{
    static char buff[256];
    
    switch (medium_unit_byte & 0x3F)
    {
        case 0x00:    
            snprintf(buff, sizeof(buff), "h,m,s");
            break;    
        case 0x01:    
            snprintf(buff, sizeof(buff), "D,M,Y");
            break;    
            
        case 0x02:    
            snprintf(buff, sizeof(buff), "Wh");
            break;    
        case 0x03:    
            snprintf(buff, sizeof(buff), "10 Wh");
            break;    
        case 0x04:    
            snprintf(buff, sizeof(buff), "100 Wh");
            break;    
        case 0x05:    
            snprintf(buff, sizeof(buff), "kWh");
            break;    
        case 0x06:    
            snprintf(buff, sizeof(buff), "10 kWh");
            break;    
        case 0x07:    
            snprintf(buff, sizeof(buff), "100 kWh");
            break;    
        case 0x08:    
            snprintf(buff, sizeof(buff), "MWh");
            break;    
        case 0x09:    
            snprintf(buff, sizeof(buff), "10 MWh");
            break;    
        case 0x0A:    
            snprintf(buff, sizeof(buff), "100 MWh");
            break;    
            
        case 0x0B:    
            snprintf(buff, sizeof(buff), "kJ");
            break;    
        case 0x0C:    
            snprintf(buff, sizeof(buff), "10 kJ");
            break;    
        case 0x0E:    
            snprintf(buff, sizeof(buff), "100 kJ");
            break;    
        case 0x0D:    
            snprintf(buff, sizeof(buff), "MJ");
            break;    
        case 0x0F:    
            snprintf(buff, sizeof(buff), "10 MJ");
            break;    
        case 0x10:    
            snprintf(buff, sizeof(buff), "100 MJ");
            break;
        case 0x11:    
            snprintf(buff, sizeof(buff), "GJ");
            break;    
        case 0x12:    
            snprintf(buff, sizeof(buff), "10 GJ");
            break;    
        case 0x13:    
            snprintf(buff, sizeof(buff), "100 GJ");
            break;

        case 0x14:    
            snprintf(buff, sizeof(buff), "W");
            break;    
        case 0x15:    
            snprintf(buff, sizeof(buff), "10 W");
            break;    
        case 0x16:    
            snprintf(buff, sizeof(buff), "100 W");
            break;    
        case 0x17:    
            snprintf(buff, sizeof(buff), "kW");
            break;    
        case 0x18:    
            snprintf(buff, sizeof(buff), "10 kW");
            break;    
        case 0x19:    
            snprintf(buff, sizeof(buff), "100 kW");
            break;
        case 0x1A:    
            snprintf(buff, sizeof(buff), "MW");
            break;    
        case 0x1B:    
            snprintf(buff, sizeof(buff), "10 MW");
            break;    
        case 0x1C:    
            snprintf(buff, sizeof(buff), "100 MW");
            break;
            
        case 0x29:    
            snprintf(buff, sizeof(buff), "l");
            break;
        case 0x2A:    
            snprintf(buff, sizeof(buff), "10 l");
            break;
        case 0x2B:    
            snprintf(buff, sizeof(buff), "100 l");
            break;
        case 0x3E:    
            snprintf(buff, sizeof(buff), "same but historic");
            break;  
        default:
            snprintf(buff, sizeof(buff), "unknown");
            break;    
    }

    return buff;
}

//------------------------------------------------------------------------------
//
// VARIABLE-LENGTH DATA RECORD FUNCTIONS
//
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//
// Medium                                                              Code bin    Code hex
// Other                                                              0000 0000        00
// Oil                                                                0000 0001        01
// Electricity                                                        0000 0010        02
// Gas                                                                0000 0011        03
// Heat (Volume measured at return temperature: outlet)               0000 0100        04
// Steam                                                              0000 0101        05
// Hot Water                                                          0000 0110        06
// Water                                                              0000 0111        07
// Heat Cost Allocator.                                               0000 1000        08
// Compressed Air                                                     0000 1001        09
// Cooling load meter (Volume measured at return temperature: outlet) 0000 1010        0A
// Cooling load meter (Volume measured at flow temperature: inlet) ♣  0000 1011        0B
// Heat (Volume measured at flow temperature: inlet)                  0000 1100        0C
// Heat / Cooling load meter ♣                                        0000 1101        OD
// Bus / System                                                       0000 1110        0E
// Unknown Medium                                                     0000 1111        0F
// Reserved                                                           .......... 10 to 15
// Cold Water                                                         0001 0110        16
// Dual Water                                                         0001 0111        17
// Pressure                                                           0001 1000        18
// A/D Converter                                                      0001 1001        19
// Reserved                                                           .......... 20 to FF
//------------------------------------------------------------------------------
const char *
mbus_data_variable_medium_lookup(u_char medium)
{
    static char buff[256];

    switch (medium)
    {
        case MBUS_VARIABLE_DATA_MEDIUM_OTHER:
            snprintf(buff, sizeof(buff), "Other");
            break;
                                          
        case MBUS_VARIABLE_DATA_MEDIUM_OIL:
            snprintf(buff, sizeof(buff), "Oil");
            break;
                  
        case MBUS_VARIABLE_DATA_MEDIUM_ELECTRICITY:
            snprintf(buff, sizeof(buff), "Electricity");
            break;
                                          
        case MBUS_VARIABLE_DATA_MEDIUM_GAS:
            snprintf(buff, sizeof(buff), "Gas");
            break;
                                          
        case MBUS_VARIABLE_DATA_MEDIUM_HEAT:
            snprintf(buff, sizeof(buff), "Heat");
            break;
                                          
        // add more ...
        default:
            snprintf(buff, sizeof(buff), "Unknown");
            break;            
    }

    return buff;       
}

//------------------------------------------------------------------------------
// Lookup the unit description from a VIF field in a data record
//------------------------------------------------------------------------------
const char *
mbus_unit_prefix(int exp)
{
    static char buff[256];
    
    switch (exp)
    {
        case 0:
            buff[0] = 0;
            break;

        case -3:
            snprintf(buff, sizeof(buff), "m");
            break;
       
        case -6:
            snprintf(buff, sizeof(buff), "my");
            break;

        case 1:
            snprintf(buff, sizeof(buff), "10 ");
            break;

        case 2:
            snprintf(buff, sizeof(buff), "100 ");
            break;

        case 3:
            snprintf(buff, sizeof(buff), "k");
            break;
            
        case 4:
            snprintf(buff, sizeof(buff), "10 k");
            break;            

        case 5:
            snprintf(buff, sizeof(buff), "100 k");
            break;            

        case 6:
            snprintf(buff, sizeof(buff), "M");
            break;
       
        case 9:
            snprintf(buff, sizeof(buff), "T");
            break;
           
        default:
            snprintf(buff, sizeof(buff), "10^%d ", exp);
    }
        
    return buff;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
const char *
mbus_vif_unit_lookup(u_char vif)
{
    static char buff[256];
    int n;

    switch (vif & 0x78)
    {
        case 0x00:
            n = (vif & 0x07) - 3;
            snprintf(buff, sizeof(buff), "Energy (%sWh)", mbus_unit_prefix(n)); 
            break;

        case 0x08:
            // 0000 1nnn          Energy       10(nnn)J     (0.001kJ to 10000kJ)
            
            n = (vif & 0x07);
            snprintf(buff, sizeof(buff), "Energy (%sJ)", mbus_unit_prefix(n));
  
            break;        

        case 0x10:

            n = (vif & 0x07);
            snprintf(buff, sizeof(buff), "Volume (%s m^3)", mbus_unit_prefix(n-6));
  
            break;

        case 0x28:

            n = (vif & 0x07);
            snprintf(buff, sizeof(buff), "Power (%sW)", mbus_unit_prefix(n-3));
            //snprintf(buff, sizeof(buff), "Power (10^%d W)", n-3);
  
            break;
            
        case 0x38:

            n = (vif & 0x07);
            snprintf(buff, sizeof(buff), "Volume flow (%s m^3/h)", mbus_unit_prefix(n-6));
  
            break;       

        case 0x58:

            n = (vif & 0x03);
            if (vif & 0x4)
                snprintf(buff, sizeof(buff), "Return temperature (%sdeg C)", mbus_unit_prefix(n-3));
            else
                snprintf(buff, sizeof(buff), "Flow temperature (%sdeg C)", mbus_unit_prefix(n-3));
  
            break;

        case 0x20:
            {
                int offset;

                if (vif & 0x4)
                    offset = snprintf(buff, sizeof(buff), "Operating time ");
                else
                    offset = snprintf(buff, sizeof(buff), "On time ");
                
                switch (vif & 0x03)
                {
                    case 0x00:
                        snprintf(&buff[offset], sizeof(buff)-offset, "(seconds)");
                        break;
                    case 0x01:
                        snprintf(&buff[offset], sizeof(buff)-offset, "(minutes)");
                        break;
                    case 0x02:
                        snprintf(&buff[offset], sizeof(buff)-offset, "(hours)");
                        break;
                    case 0x03:
                        snprintf(&buff[offset], sizeof(buff)-offset, "(days)");
                        break;
                }
            }
            break;                     

        case 0x68:

            if (vif & 0x1)
                snprintf(buff, sizeof(buff), "Time Point (time & date)");
            else
                snprintf(buff, sizeof(buff), "Time Point (date)");
  
            break;
            
        case 0x60:
        // E110 00nn    Temperature Difference   10(nn-3)K   (mK to  K)
        // E110 01nn    External Temperature     10(nn-3)°C  (0.001°C to 1°C)

            n = (vif & 0x03);
            
            if (vif & 0x4)
                snprintf(buff, sizeof(buff), "External Temperature (%s deg C)", mbus_unit_prefix(n-3));
            else
                snprintf(buff, sizeof(buff), "Temperature Difference (%s deg C)", mbus_unit_prefix(n-3));
                        
            break;
                                       
        default:
        
            snprintf(buff, sizeof(buff), "Unknown (VIF=0x%.2X)", vif);
            break;
    }
    

    return buff;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
const char *
mbus_vib_unit_lookup(mbus_value_information_block *vib)
{   
    static char buff[256];
    u_char vif;
    int n;

    if ((vib->vif & 0xFF) == 0xFD)
    {
        if (vib->nvife == 0)
        {
            snprintf(buff, sizeof(buff), "Missing VIF extension");
        }
        else if ((vib->vife[0] & 0xFF) == (0x10))
        {
            snprintf(buff, sizeof(buff), "Customer location");
        }
        else
        {
            snprintf(buff, sizeof(buff), "Unrecongized VIF extension: 0x%x", vib->vife[0] & 0xFF);
        }
        return buff;
    }
    
    vif = vib->vif;

    return mbus_vif_unit_lookup(vif);    
}

//------------------------------------------------------------------------------
// Decode data and write to string
//
// Data format (for record->data data array)
// 
// Length in Bit   Code    Meaning           Code      Meaning
//      0          0000    No data           1000      Selection for Readout
//      8          0001     8 Bit Integer    1001      2 digit BCD
//     16          0010    16 Bit Integer    1010      4 digit BCD
//     24          0011    24 Bit Integer    1011      6 digit BCD
//     32          0100    32 Bit Integer    1100      8 digit BCD
//   32 / N        0101    32 Bit Real       1101      variable length
//     48          0110    48 Bit Integer    1110      12 digit BCD
//     64          0111    64 Bit Integer    1111      Special Functions
//
// The Code is stored in record->drh.dib.dif
//
// Return a string containing the data
//
// Source: MBDOC48.PDF
//
//------------------------------------------------------------------------------
const char *
mbus_data_record_decode(mbus_data_record *record)
{
    static char buff[256];

    if (record)
    {
        int val;
            
        switch (record->drh.dib.dif & 0x0F)
        {
            case 0x00: // no data
        
                buff[0] = 0;
                
                break; 

            case 0x01: // 1 byte integer (8 bit)
        
                val = mbus_data_int_decode(record->data, 1);
               
                snprintf(buff, sizeof(buff), "%d", val);

                //snprintf(buff, sizeof(buff), "%d", (int)(record->data[0]));
                break; 


            case 0x02: // 2 byte integer (16 bit)
        
                //val =              record->data[1];
                //val = (val << 8) + record->data[0];
        
                val = mbus_data_int_decode(record->data, 2);
        
                snprintf(buff, sizeof(buff), "%d", val);

                break; 

            case 0x03: // 3 byte integer (24 bit)
        
                //val =              record->data[2];
                //val = (val << 8) + record->data[1];
                //val = (val << 8) + record->data[0];

                val = mbus_data_int_decode(record->data, 3);
                        
                snprintf(buff, sizeof(buff), "%d", val);

                break; 
                
            case 0x04: // 4 byte integer (32 bit)
        
                //val =              (record->data[3]&0xFF);
                //val = (val << 8) | (record->data[2]&0xFF);
                //val = (val << 8) | (record->data[1]&0xFF);
                //val = (val << 8) | (record->data[0]&0xFF);

                val = mbus_data_int_decode(record->data, 4);
        
                snprintf(buff, sizeof(buff), "%d", val);

                break;          

            case 0x09: // 2 digit BCD (8 bit)
  
                val = mbus_data_bcd_decode(record->data, 1);  
                snprintf(buff, sizeof(buff), "%d", val);     
        
//                snprintf(buff, sizeof(buff), "%d%d",
//                         (record->data[0]>>4)&0x0F, (record->data[0])&0x0F);
                break;
                
            case 0x0A: // 4 digit BCD (16 bit)
        
                val = mbus_data_bcd_decode(record->data, 2);  
                snprintf(buff, sizeof(buff), "%d", val);     
                
//                snprintf(buff, sizeof(buff), "%d%d%d%d",
//                         (record->data[1]>>4)&0x0F, (record->data[1])&0x0F, 
//                         (record->data[0]>>4)&0x0F, (record->data[0])&0x0F);

                break;

            case 0x0B: // 6 digit BCD (24 bit)

                val = mbus_data_bcd_decode(record->data, 3);  
                snprintf(buff, sizeof(buff), "%d", val);     
        
//                snprintf(buff, sizeof(buff), "%d%d%d%d%d%d",
//                         (record->data[2]>>4)&0x0F, (record->data[2])&0x0F, 
//                         (record->data[1]>>4)&0x0F, (record->data[1])&0x0F, 
//                         (record->data[0]>>4)&0x0F, (record->data[0])&0x0F);

                break;
                
            case 0x0C: // 8 digit BCD (32 bit)

                val = mbus_data_bcd_decode(record->data, 4);  
                snprintf(buff, sizeof(buff), "%d", val);     
        
//                snprintf(buff, sizeof(buff), "%d%d%d%d%d%d%d%d",
//                         (record->data[3]>>4)&0x0F, (record->data[3])&0x0F, 
//                         (record->data[2]>>4)&0x0F, (record->data[2])&0x0F, 
//                         (record->data[1]>>4)&0x0F, (record->data[1])&0x0F, 
//                         (record->data[0]>>4)&0x0F, (record->data[0])&0x0F);

                break;
                                                                  
            default:
        
                snprintf(buff, sizeof(buff), "Unknown");
                break;
        }

        return buff;     
    }

    return NULL;
}
//------------------------------------------------------------------------------
// Return the unit description for a variable-length data record
//------------------------------------------------------------------------------
const char *
mbus_data_record_unit(mbus_data_record *record)
{
    static char buff[128];
    
    if (record)
    {
        //snprintf(buff, sizeof(buff), "%s", mbus_vif_unit_lookup(record->drh.vib.vif));
        snprintf(buff, sizeof(buff), "%s", mbus_vib_unit_lookup(&(record->drh.vib)));
                 
        return buff;    
    }
    
    return NULL;
}

//------------------------------------------------------------------------------
// Return the numerical value for a variable-length data record
//------------------------------------------------------------------------------
const char *
mbus_data_record_value(mbus_data_record *record)
{
    static char buff[128];
    
    if (record)
    {

        snprintf(buff, sizeof(buff), "%s", mbus_data_record_decode(record));
                 
        return buff;    
    }
    
    return NULL;
}

//------------------------------------------------------------------------------
// Return a string containing the function description
//------------------------------------------------------------------------------
const char *
mbus_data_record_function(mbus_data_record *record)
{
    static char buff[128];
    
    if (record)
    {
        switch (record->drh.dib.dif & 0x30)
        {
            case 0x00:
                snprintf(buff, sizeof(buff), "Instantaneous value");           
                break;

            case 0x10:
                snprintf(buff, sizeof(buff), "Minimum value");                      
                break;

            case 0x20:
                snprintf(buff, sizeof(buff), "Maximum value");                       
                break;

            case 0x30:
                snprintf(buff, sizeof(buff), "Value during error state");                       
                break;
        
            default:
                snprintf(buff, sizeof(buff), "unknown");                 
        }

        return buff;    
    }
    
    return NULL;
}

const char *
mbus_data_fixed_function(int status)
{
    static char buff[128];

    snprintf(buff, sizeof(buff), "%s",
            (status & MBUS_DATA_FIXED_STATUS_DATE_MASK) == MBUS_DATA_FIXED_STATUS_DATE_STORED ? "Stored value" : "Instantaneous value" );
    
    return buff;
}

//------------------------------------------------------------------------------
//
// PARSER FUNCTIONS
//
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// PARSE M-BUS frame data structures from binary data.
//------------------------------------------------------------------------------
int
mbus_parse(mbus_frame *frame, u_char *data, size_t data_size)
{
    size_t i, len;
    
    if (frame && data && data_size > 0)
    {
        //printf("%s: Attempting to parse binary data [size = %d]\n", __PRETTY_FUNCTION__, data_size);
        //printf("%s: ", __PRETTY_FUNCTION__);
        //for (i = 0; i < data_size; i++)
        //{
        //    printf("%.2X ", data[i] & 0xFF);
        //}
        //printf("\n%s: done.\n", __PRETTY_FUNCTION__);
    
        switch (data[0] & 0xFF)
        {
            case MBUS_FRAME_ACK_START:
            
                // OK, got a valid ack frame, require no more data 
                frame->start1   = data[0];
                frame->type = MBUS_FRAME_TYPE_ACK;
                return 0;
                //return MBUS_FRAME_BASE_SIZE_ACK - 1; // == 0
                              
            case MBUS_FRAME_SHORT_START:
            
                if (data_size < MBUS_FRAME_BASE_SIZE_SHORT)
                {
                    // OK, got a valid short packet start, but we need more data
                    return MBUS_FRAME_BASE_SIZE_SHORT - data_size;
                }
                
                if (data_size != MBUS_FRAME_BASE_SIZE_SHORT)
                {
                    // too much data... ?
                    return -2;
                }
                
                // init frame data structure
                frame->start1   = data[0];
                frame->control  = data[1];
                frame->address  = data[2];
                frame->checksum = data[3];
                frame->stop     = data[4];
                
                frame->type = MBUS_FRAME_TYPE_SHORT;

                // verify the frame
                if (mbus_frame_verify(frame) != 0)
                {
                    return -3;
                }

                // successfully parsed data
                return 0;

            case MBUS_FRAME_LONG_START: // (also CONTROL)
            
                if (data_size < 3)
                {
                    // OK, got a valid long/control packet start, but we need
                    // more data to determine the length
                    return 3 - data_size;
                }
                
                // init frame data structure
                frame->start1   = data[0];
                frame->length1  = data[1];
                frame->length2  = data[2];
                
                if (frame->length1 != frame->length2)
                {
                    // not a valid M-bus frame
                    return -2;
                }
                
                // check length of packet:
                len = frame->length1;
                
                if (data_size < (size_t)(MBUS_FRAME_FIXED_SIZE_LONG + len))
                {
                    // OK, but we need more data
                    return MBUS_FRAME_FIXED_SIZE_LONG + len - data_size;                
                }
                
                // we got the whole packet, continue parsing
                frame->start2   = data[3];
                frame->control  = data[4];
                frame->address  = data[5];
                frame->control_information = data[6];
                
                frame->data_size = len - 3;
                for (i = 0; i < frame->data_size; i++)
                {
                    frame->data[i] = data[7 + i];
                }                              
                                
                frame->checksum = data[data_size-2]; // data[6 + frame->data_size + 1]
                frame->stop     = data[data_size-1]; // data[6 + frame->data_size + 2]          
                
                if (frame->data_size == 0)
                {
                    frame->type = MBUS_FRAME_TYPE_CONTROL;
                }
                else
                {
                    frame->type = MBUS_FRAME_TYPE_LONG;                
                }

                // verify the frame
                if (mbus_frame_verify(frame) != 0)
                {
                    return -3;
                }

                // successfully parsed data
                return 0;        
            default:
                // not a valid M-Bus frame header (start byte)
                return -4;
        }    
        
    }

    return -1;
}


//------------------------------------------------------------------------------
// Parse the fixed-length data of a M-Bus frame
//------------------------------------------------------------------------------
int
mbus_data_fixed_parse(mbus_frame *frame, mbus_data_fixed *data)
{
    if (frame && data)
    {  
        // copy the fixed-length data structure
        memcpy((void *)data, (void *)(frame->data), sizeof(mbus_data_fixed));

        return 0;
    }    

    return -1;
}


//------------------------------------------------------------------------------
// Parse the variable-length data of a M-Bus frame
//------------------------------------------------------------------------------
int
mbus_data_variable_parse(mbus_frame *frame, mbus_data_variable *data)
{
    size_t i, j;
    
    if (frame && data)
    {
        data->nrecords = 0;
        
        // first parse the variable data fixed header  
        memcpy((void *)&(data->header), (void *)(frame->data), sizeof(mbus_data_variable_header));
        i = 12;    
               
        while (i < frame->data_size)
        {
            // read and parse DIB
        
            // DIF
            data->record[data->nrecords].drh.dib.dif = frame->data[i];

            if (data->record[data->nrecords].drh.dib.dif == 0x0F) //MBUS_DIB_DIF_VENDOR_SPECIFIC)
            {
                i++;
                // copy the remaining data as it is vendor specific
                data->record[data->nrecords].data_len = frame->data_size - i;
                for (j = 0; j < data->record[data->nrecords].data_len; j++)
                {
                    data->record[data->nrecords].data[j] = frame->data[i++];
                }
                continue;
            }
        
            // calculate length of data record
            data->record[data->nrecords].data_len = data->record[data->nrecords].drh.dib.dif & 0x07;

            // read DIF extensions
            data->record[data->nrecords].drh.dib.ndife = 0;
            while (frame->data[i] & MBUS_DIB_DIF_EXTENSION_BIT && data->record[data->nrecords].drh.dib.ndife < NITEMS(data->record[data->nrecords].drh.dib.dife))
            {
                u_char dife = frame->data[i+1];
                data->record[data->nrecords].drh.dib.dife[data->record[data->nrecords].drh.dib.ndife] = dife;

                data->record[data->nrecords].drh.dib.ndife++;
                i++;
            }
            i++;

            // VIF
            data->record[data->nrecords].drh.vib.vif = frame->data[i]; 
                
            // VIFE
            data->record[data->nrecords].drh.vib.nvife = 0;
            while (frame->data[i] & MBUS_DIB_VIF_EXTENSION_BIT && data->record[data->nrecords].drh.vib.nvife < NITEMS(data->record[data->nrecords].drh.vib.vife))
            {
                u_char vife = frame->data[i+1];
                data->record[data->nrecords].drh.vib.vife[data->record[data->nrecords].drh.vib.nvife] = vife;
            
                data->record[data->nrecords].drh.vib.nvife++;
                i++;
            }
            i++;
                    
            // copy data
            for (j = 0; j < data->record[data->nrecords].data_len; j++)
            {
                data->record[data->nrecords].data[j] = frame->data[i++];       
            }

            // move on to next data record
            data->nrecords++;
        }
        
        return 0;
    }    
    
    return -1;
}

//------------------------------------------------------------------------------
// Check the stype of the frame data (fixed or variable) and dispatch to the
// corresponding parser function.
//------------------------------------------------------------------------------
int
mbus_frame_data_parse(mbus_frame *frame, mbus_frame_data *data)
{
    if (frame && data && frame->data_size > 0)
    {
        if (frame->control_information == MBUS_CONTROL_INFO_RESP_FIXED)
        {
            data->type = MBUS_DATA_TYPE_FIXED;
            return mbus_data_fixed_parse(frame, &(data->data_fix));
        }
        
        if (frame->control_information == MBUS_CONTROL_INFO_RESP_VARIABLE)
        {
            data->type = MBUS_DATA_TYPE_VARIABLE;
            return mbus_data_variable_parse(frame, &(data->data_var));
        }
    }
    
    return -1;
}

//------------------------------------------------------------------------------
// Pack the M-bus frame into a binary string representation that can be sent 
// on the bus. The binary packet format is different for the different types
// of M-bus frames.
//------------------------------------------------------------------------------
int
mbus_frame_pack(mbus_frame *frame, u_char *data, size_t data_size)
{
    size_t i, offset = 0;
    
    if (frame && data)
    {   
        if (mbus_frame_calc_length(frame) == -1)
        {
            return -2;
        }

        if (mbus_frame_calc_checksum(frame) == -1)
        {
            return -3;
        }
        
        switch (frame->type)
        {
            case MBUS_FRAME_TYPE_ACK:
            
                if (data_size < MBUS_FRAME_ACK_BASE_SIZE)
                {
                    return -4;
                }
        
                data[offset++] = frame->start1; 
        
                return offset;        

            case MBUS_FRAME_TYPE_SHORT:

                if (data_size < MBUS_FRAME_SHORT_BASE_SIZE)
                {
                    return -4;
                }
                
                data[offset++] = frame->start1;
                data[offset++] = frame->control;
                data[offset++] = frame->address;
                data[offset++] = frame->checksum;
                data[offset++] = frame->stop;
        
                return offset;        

            case MBUS_FRAME_TYPE_CONTROL:
        
                if (data_size < MBUS_FRAME_CONTROL_BASE_SIZE)
                {
                    return -4;
                }
                
                data[offset++] = frame->start1;
                data[offset++] = frame->length1;
                data[offset++] = frame->length2;
                data[offset++] = frame->start2;

                data[offset++] = frame->control;
                data[offset++] = frame->address;
                data[offset++] = frame->control_information;

                data[offset++] = frame->checksum;
                data[offset++] = frame->stop;    
        
                return offset;        

            case MBUS_FRAME_TYPE_LONG:

                if (data_size < frame->data_size + MBUS_FRAME_LONG_BASE_SIZE)
                {
                    return -4;
                }

                data[offset++] = frame->start1;
                data[offset++] = frame->length1;
                data[offset++] = frame->length2;
                data[offset++] = frame->start2;

                data[offset++] = frame->control;
                data[offset++] = frame->address;
                data[offset++] = frame->control_information;

                for (i = 0; i < frame->data_size; i++)
                {
                    data[offset++] = frame->data[i];
                }
                
                data[offset++] = frame->checksum;
                data[offset++] = frame->stop;    
        
                return offset;      
            
            default:
                return -5;
        }        
    }

    return -1;
}


//------------------------------------------------------------------------------
// pack the data stuctures into frame->data
//------------------------------------------------------------------------------
int
mbus_frame_internal_pack(mbus_frame *frame, mbus_frame_data *frame_data)
{
    int i, j;

    if (frame == NULL || frame_data == NULL)
        return -1;

    frame->data_size = 0;

    switch (frame_data->type)
    {
        case MBUS_DATA_TYPE_FIXED:

            //
            // pack fixed data structure
            //
            frame->data[frame->data_size++] = frame_data->data_fix.id_bcd[0];
            frame->data[frame->data_size++] = frame_data->data_fix.id_bcd[1];
            frame->data[frame->data_size++] = frame_data->data_fix.id_bcd[2];
            frame->data[frame->data_size++] = frame_data->data_fix.id_bcd[3];
            frame->data[frame->data_size++] = frame_data->data_fix.tx_cnt;
            frame->data[frame->data_size++] = frame_data->data_fix.status;
            frame->data[frame->data_size++] = frame_data->data_fix.cnt1_type;
            frame->data[frame->data_size++] = frame_data->data_fix.cnt2_type;
            frame->data[frame->data_size++] = frame_data->data_fix.cnt1_val[0];
            frame->data[frame->data_size++] = frame_data->data_fix.cnt1_val[1];
            frame->data[frame->data_size++] = frame_data->data_fix.cnt1_val[2];
            frame->data[frame->data_size++] = frame_data->data_fix.cnt1_val[3];
            frame->data[frame->data_size++] = frame_data->data_fix.cnt2_val[0];
            frame->data[frame->data_size++] = frame_data->data_fix.cnt2_val[1];
            frame->data[frame->data_size++] = frame_data->data_fix.cnt2_val[2];
            frame->data[frame->data_size++] = frame_data->data_fix.cnt2_val[3];

            break;

        case MBUS_DATA_TYPE_VARIABLE:

            //
            // first pack variable data structure header
            //
            frame->data[frame->data_size++] = frame_data->data_var.header.id_bcd[0];
            frame->data[frame->data_size++] = frame_data->data_var.header.id_bcd[1];
            frame->data[frame->data_size++] = frame_data->data_var.header.id_bcd[2];
            frame->data[frame->data_size++] = frame_data->data_var.header.id_bcd[3];
            frame->data[frame->data_size++] = frame_data->data_var.header.manufacturer[0];
            frame->data[frame->data_size++] = frame_data->data_var.header.manufacturer[1];
            frame->data[frame->data_size++] = frame_data->data_var.header.version;
            frame->data[frame->data_size++] = frame_data->data_var.header.medium;
            frame->data[frame->data_size++] = frame_data->data_var.header.access_no;
            frame->data[frame->data_size++] = frame_data->data_var.header.status;
            frame->data[frame->data_size++] = frame_data->data_var.header.signature[0];
            frame->data[frame->data_size++] = frame_data->data_var.header.signature[1];

            //
            // pack all data records
            //
            for (i = 0; i < frame_data->data_var.nrecords; i++)
            {
                // pack DIF
                //printf("%s: parsing DIF [%d]", __PRETTY_FUNCTION__, frame->data_size);
                frame->data[frame->data_size++] = frame_data->data_var.record[i].drh.dib.dif;
                for (j = 0; j < frame_data->data_var.record[i].drh.dib.ndife; j++)
                {
                    frame->data[frame->data_size++] = frame_data->data_var.record[i].drh.dib.dife[j];
                }

                // pack VIF
                //printf("%s: parsing VIF [%d]", __PRETTY_FUNCTION__, frame->data_size);
                frame->data[frame->data_size++] = frame_data->data_var.record[i].drh.vib.vif;
                for (j = 0; j < frame_data->data_var.record[i].drh.vib.nvife; j++)
                {
                    frame->data[frame->data_size++] = frame_data->data_var.record[i].drh.vib.vife[j];
                }

                // pack data
                //printf("%s: parsing data [%d : %d]", __PRETTY_FUNCTION__, frame->data_size, frame_data->data_var.record[i].data_len);
                for (j = 0; j < frame_data->data_var.record[i].data_len; j++)
                {
                    frame->data[frame->data_size++] = frame_data->data_var.record[i].data[j];
                }
            }         
            
            break;

        default:
            return -2;
    }

    return 0;
}

//------------------------------------------------------------------------------
//
// Print/Dump functions
//
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
// Dump frame in HEX on standard output
//------------------------------------------------------------------------------
int
mbus_frame_print(mbus_frame *frame)
{
    u_char data_buff[256];
    int len, i;
    
    if (frame)
    {
        if ((len = mbus_frame_pack(frame, data_buff, sizeof(data_buff))) == -1)
        {
            return -2;
        }
       
        printf("%s: Dumping M-Bus frame [type %d, %d bytes]: ", __PRETTY_FUNCTION__, frame->type, len);
        for (i = 0; i < len; i++)
        {
            printf("%.2X ", data_buff[i] & 0xFF);
        }
        printf("\n");
        
        return 0;
    }
    
    return -1;
}

int
mbus_frame_data_print(mbus_frame_data *data)
{
    if (data)
    {
        if (data->type == MBUS_DATA_TYPE_FIXED)
        {
            return mbus_data_fixed_print(&(data->data_fix));
        }
        
        if (data->type == MBUS_DATA_TYPE_VARIABLE)
        {
            return mbus_data_variable_print(&(data->data_var));
        }
    }
    
    return -1;
}

//------------------------------------------------------------------------------
// Print M-bus frame info to stdout
//------------------------------------------------------------------------------
int
mbus_data_variable_header_print(mbus_data_variable_header *header)
{
    if (header)
    {
        printf("%s: ID           = %d\n", __PRETTY_FUNCTION__, mbus_data_bcd_decode(header->id_bcd, 4));

        printf("%s: Manufacturer = 0x%.2X%.2X\n", __PRETTY_FUNCTION__,
               header->manufacturer[1] & 0xFF, header->manufacturer[0] & 0xFF);
               
        printf("%s: Manufacturer = %s\n", __PRETTY_FUNCTION__,
               mbus_decode_manufacturer(header->manufacturer[0], header->manufacturer[1]));               

        printf("%s: Version      = 0x%.2X\n", __PRETTY_FUNCTION__, header->version);
        printf("%s: Medium       = %s (0x%.2X)\n", __PRETTY_FUNCTION__, mbus_data_variable_medium_lookup(header->medium), header->medium);
        printf("%s: Access #     = 0x%.2X\n", __PRETTY_FUNCTION__, header->access_no);
        printf("%s: Status       = 0x%.2X\n", __PRETTY_FUNCTION__, header->status);
        printf("%s: Signature    = 0x%.2X%.2X\n", __PRETTY_FUNCTION__,
               header->signature[0], header->signature[1]);

    }
    
    return -1;
}

int
mbus_data_variable_print(mbus_data_variable *data)
{
    size_t i, j;
    
    if (data)
    {   
        mbus_data_variable_header_print(&(data->header));
    
        for (i = 0; i < data->nrecords; i++)
        {
            // DIF
            printf("DIF           = %.2X\n", data->record[i].drh.dib.dif & 0xFF);
            printf("DIF.Extension = %s\n",   (data->record[i].drh.dib.dif & MBUS_DIB_DIF_EXTENSION_BIT) ? "Yes":"No");        
            printf("DIF.Function  = %s\n",   (data->record[i].drh.dib.dif & 0x30) ? "Minimum value" : "Instantaneous value" );
            printf("DIF.Data      = %.2X\n", data->record[i].drh.dib.dif & 0x0F);

            // VENDOR SPECIFIC
            if (data->record[i].drh.dib.dif == 0x0F) //MBUS_DIB_DIF_VENDOR_SPECIFIC)
            {
                printf("%s: DEBUG: VENDOR DATA [size=%zd] = ", __PRETTY_FUNCTION__, data->record[i].data_len);
                for (j = 0; j < data->record[i].data_len; j++)
                {
                    printf("%.2X ", data->record[i].data[j] & 0xFF);        
                }
                printf("\n");
                continue;
            }
        
            // calculate length of data record
            printf("DATA LENGTH = %zd\n", data->record[i].data_len);

            // DIFE
            for (j = 0; j < data->record[i].drh.dib.ndife; j++)
            {
                u_char dife = data->record[i].drh.dib.dife[j];
                
                printf("DIFE[%zd]           = %.2X\n", j,  dife & 0xFF);
                printf("DIFE[%zd].Extension = %s\n",   j, (dife & MBUS_DIB_DIF_EXTENSION_BIT) ? "Yes" : "No");        
                printf("DIFE[%zd].Function  = %s\n",   j, (dife & 0x30) ? "Minimum value" : "Instantaneous value" );
                printf("DIFE[%zd].Data      = %.2X\n", j,  dife & 0x0F);            
            }
   
        }
    }
    
    return -1;
}

int
mbus_data_fixed_print(mbus_data_fixed *data)
{
    if (data)
    {    
        printf("%s: ID       = %d\n", __PRETTY_FUNCTION__, mbus_data_bcd_decode(data->id_bcd, 4));
        printf("%s: Access # = 0x%.2X\n", __PRETTY_FUNCTION__, data->tx_cnt);
        printf("%s: Status   = 0x%.2X\n", __PRETTY_FUNCTION__, data->status);
        printf("%s: Function = %s\n", __PRETTY_FUNCTION__, mbus_data_fixed_function(data->status));
        
        printf("%s: Medium1  = %s\n", __PRETTY_FUNCTION__, mbus_data_fixed_medium(data));
        printf("%s: Unit1    = %s\n", __PRETTY_FUNCTION__, mbus_data_fixed_unit(data->cnt1_type));
        if ((data->status & MBUS_DATA_FIXED_STATUS_FORMAT_MASK) == MBUS_DATA_FIXED_STATUS_FORMAT_BCD)
        {
            printf("%s: Counter1 = %d\n", __PRETTY_FUNCTION__, mbus_data_bcd_decode(data->cnt1_val, 4));
        }
        else
        {
            printf("%s: Counter1 = %d\n", __PRETTY_FUNCTION__, mbus_data_int_decode(data->cnt1_val, 4));
        }

        printf("%s: Medium2  = %s\n", __PRETTY_FUNCTION__, mbus_data_fixed_medium(data));
        printf("%s: Unit2    = %s\n", __PRETTY_FUNCTION__, mbus_data_fixed_unit(data->cnt2_type));
        if ((data->status & MBUS_DATA_FIXED_STATUS_FORMAT_MASK) == MBUS_DATA_FIXED_STATUS_FORMAT_BCD)
        {
            printf("%s: Counter2 = %d\n", __PRETTY_FUNCTION__, mbus_data_bcd_decode(data->cnt2_val, 4));
        }
        else
        {
            printf("%s: Counter2 = %d\n", __PRETTY_FUNCTION__, mbus_data_int_decode(data->cnt2_val, 4));        
        }          
    }
    
    return -1;
}

//------------------------------------------------------------------------------
//
// XML RELATED FUNCTIONS
//
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Generate XML for the variable-length data header
//------------------------------------------------------------------------------
int
mbus_data_variable_header_print_xml(mbus_data_variable_header *header)
{
    int val;
    
    if (header)
    {
        printf("    <SlaveInformation>\n");

        val = mbus_data_bcd_decode(header->id_bcd, 4);
        
        printf("        <Id>%d</Id>\n", val);
        printf("        <Manufacturer>%s</Manufacturer>\n",
                mbus_decode_manufacturer(header->manufacturer[0], header->manufacturer[1]));               
        printf("        <Version>%d</Version>\n", header->version);
        printf("        <Medium>%s</Medium>\n",   mbus_data_variable_medium_lookup(header->medium));
        printf("        <AccessNumber>%d</AccessNumber>\n", header->access_no);
        printf("        <Status>%.2X</Status>\n", header->status);
        printf("        <Signature>%.2X%.2X</Signature>\n", header->signature[0], header->signature[1]);

        printf("    </SlaveInformation>\n\n");
    }
    
    return -1;
}

//------------------------------------------------------------------------------
// Generate XML for variable-length data 
//------------------------------------------------------------------------------
int
mbus_data_variable_print_xml(mbus_data_variable *data)
{
    size_t i;
    
    if (data)
    {
        printf("<MBusData>\n\n");
        
        mbus_data_variable_header_print_xml(&(data->header));
    
        // data->record[data->nrecords]
        for (i = 0; i < data->nrecords; i++)
        {
            printf("    <DataRecord id=\"%zd\">\n", i);
            printf("        <Function>%s</Function>\n", mbus_data_record_function(&(data->record[i])));
            printf("        <Unit>%s</Unit>\n",         mbus_data_record_unit(&(data->record[i])));
            printf("        <Value>%s</Value>\n",       mbus_data_record_value(&(data->record[i])));
            printf("    </DataRecord>\n\n");        
        }
       
        printf("</MBusData>\n");
    }
    
    return -1;
}

int
mbus_data_fixed_print_xml(mbus_data_fixed *data)
{
    if (data)
    {
        printf("<MBusData>\n\n");
    
        printf("    <SlaveInformation>\n");
        printf("        <Id>%d</Id>\n", mbus_data_bcd_decode(data->id_bcd, 4));
        printf("        <Medium>%s</Medium>\n", mbus_data_fixed_medium(data));
        printf("        <AccessNumber>%d</AccessNumber>\n", data->tx_cnt);
        printf("        <Status>%.2X</Status>\n", data->status);
        printf("    </SlaveInformation>\n\n");
             
        printf("    <DataRecord id=\"0\">\n");
        printf("        <Function>%s</Function>\n", mbus_data_fixed_function(data->status));
        printf("        <Unit>%s</Unit>\n",         mbus_data_fixed_unit(data->cnt1_type));
        if ((data->status & MBUS_DATA_FIXED_STATUS_FORMAT_MASK) == MBUS_DATA_FIXED_STATUS_FORMAT_BCD)
        {
            printf("        <Value>%d</Value>\n", mbus_data_bcd_decode(data->cnt1_val, 4));
        }
        else
        {
            printf("        <Value>%d</Value>\n", mbus_data_int_decode(data->cnt1_val, 4));
        }
        printf("    </DataRecord>\n\n");      

        printf("    <DataRecord id=\"1\">\n");
        printf("        <Function>%s</Function>\n", mbus_data_fixed_function(data->status));
        printf("        <Unit>%s</Unit>\n",         mbus_data_fixed_unit(data->cnt2_type));
        if ((data->status & MBUS_DATA_FIXED_STATUS_FORMAT_MASK) == MBUS_DATA_FIXED_STATUS_FORMAT_BCD)
        {
            printf("        <Value>%d</Value>\n", mbus_data_bcd_decode(data->cnt2_val, 4));
        }
        else
        {
            printf("        <Value>%d</Value>\n", mbus_data_int_decode(data->cnt2_val, 4));
        }
        printf("    </DataRecord>\n\n");      

       
        printf("</MBusData>\n");
    }
    
    return -1;
}

//------------------------------------------------------------------------------
// 
//------------------------------------------------------------------------------
int
mbus_frame_data_print_xml(mbus_frame_data *data)
{
    if (data)
    {
        if (data->type == MBUS_DATA_TYPE_FIXED)
        {
            return mbus_data_fixed_print_xml(&(data->data_fix));
        }
        
        if (data->type == MBUS_DATA_TYPE_VARIABLE)
        {
            return mbus_data_variable_print_xml(&(data->data_var));
        }
    }
    
    return -1;
}

